/* drvBitBus.c */

/*
 *	Original Author: Ned Arnold
 *      Author: John Winans
 *      Date:   09-10-91
 *	XVME-402 and PB-bit BitBus driver
 */

/********************COPYRIGHT NOTIFICATION**********************************
This software is subject to the the conditions described in COPYRIGHT_Combined
and COPYRIGHT_BESSY files included as part of this distribution.
***************************************************************************/


/*
  changes to UI:

  XYCOM_BB_MAX_OUTSTAND_MSGS -> XycomMaxOutstandMsgs



  TODO:

  1) Synthetic delays after 91s.
  2) Rebuild the link kill stuff so that it does not mess with the WDtask.
  3) Verify race conditions on the slave reset function.
  4) Write a callable bug-syncer that uses the RAC get function ID command.
  5) Merge the watchdog tasks into one.
  6) Try merging the TX and RX tasks.

*/

/* Modification Log:
 * -----------------
 * .01  09-30-91        jrw     Completely redesigned and rewritten
 * .02	12-02-91	jrw	Changed priority info to arrays
 * .03	12-16-91	jrw	Made the data portion of the message a pointer
 * .04	01-21-91	jrw	moved the task parameters into task_params.h
 * .05	02-12-92	jrw	removed IRQ based transmission.
 * .06	04-08-92	jrw	moved the device configs into module_types.h
 * .07	09-28-92	jrw	upped the reset delay time to 1/2 second
 * .08	08-02-93	mrk	Added call to taskwdInsert
 * .09  06-28-94	jrw	Major work on RAC_RESET and NODE_OFFLINE stuff.
 * .10  07-01-94	jrw	Merged PEP and Xycom versions together.
 *				ANSIficated this code
 * .20  06-30-97        nda     removed alot of "debugging" code now that the
 *                              PEP board is finally behaving.
 *                              Added device support for waveforms so the
 *                              bitbus timeouts per node can be monitored 
 *                              via EPICS.
 *
 * NOTES:
 * This driver currently needs work on error message generation.
 *
 */

#include <vxWorks.h>
#include <stddef.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysLib.h>
#include <intLib.h>
#include <iosLib.h>
#include <rngLib.h>
#include <wdLib.h>
#include <tickLib.h>
#include <vme.h>
#include <iv.h>
#include <rebootLib.h>
#include <logLib.h>
#include <intLib.h>
#include <vxLib.h>

#include <epicsThread.h>
#include <epicsMutex.h>
#include <epicsEvent.h>
#include <drvSup.h>
#include <dbDefs.h>
#include <link.h>
#include <callback.h>
#include <taskwd.h>
#include <devLib.h>
#include <epicsPrint.h>
#include <menuFtype.h>


/* used by device support routines */
#include        <alarm.h>
#include        <dbAccess.h>
#include        <recSup.h>
#include        <devSup.h>
#include        <waveformRecord.h>
#include        <boRecord.h>

#include <drvBitBusInterface.h>
#include "drvBitBus.h"

#define STATIC	static
#define STATIC

#define BB_DEFAULT_RETIRE_TIME 5	/* Seconds to wait for a response */

STATIC long	reportBB(void);
STATIC long	initBB(void);
STATIC long	qBBReq(struct  dpvtBitBusHead *pdpvt, int prio);

STATIC int	xvmeReset(int link);
STATIC int	xvmeTmoHandler(int link);
void	xvmeRxTask(int link);
void	xvmeTxTask(int link);
void	xvmeWdTask(int link);
STATIC void	xvmeIrqRdav(int link);
STATIC void	xvmeIrqRcmd(int link);

STATIC int	pepReset(int link);
STATIC int	pepTmoHandler(int link);
void	pepRxTask(int link);
void	pepTxTask(int link);
void	pepWdTask(int link);
STATIC void	pepIrqRdav(int link);

STATIC int	bbKill(int link);
STATIC void	BBrebootFunc(void);

STATIC int	txStuck(int link);
/* BBConfig and BBSnoop definitions moved to drvBitBusInterface.h */

#ifdef BB_SUPER_DEBUG
int BBHistDump(int link);
STATIC int BBDumpXactHistory(XactHistStruct *pXact);
#endif

long pulseSysMon();


/*****************************************************************************
 *
 * Used to limit the TOTAL number of simultaneous messages that can be
 * outstanding on a single Xycom/PEP link.  (Shell settable.) PEP limit
 * of 7 is a magic number. Anything more and PEP board will throw away
 * your transaction resulting in a TIMEOUT. Ugh.
 *
 *****************************************************************************/
int	XycomMaxOutstandMsgs = XYCOM_BB_MAX_OUTSTAND_MSGS;
int     PepMaxOutstandMsgs = 7;  /* constraint of PEP discovered in 80C152
                                    source code */

/* On a bitbus error, if link and node match the following global variables, 
   write to the System Monitor board to generate a trigger */

int bbErrorTrigNode = 1; /* Set to -1 for all nodes */
int bbErrorTrigLink = 0;

/* Used to keep track of errors per node */
#define BB_NUM_NODES  100
#define BB_NUM_LINKS 4
unsigned short  bbNodeTMO[BB_NUM_LINKS][BB_NUM_NODES];
unsigned short  bbNode91[BB_NUM_LINKS][BB_NUM_NODES];

/*****************************************************************************
 *
 * Debugging flags that may be set from the shell.
 *
 * bbDebug	Used to get status information from the driver's internals.
 *		May be set with a value from 0 to 40.  the higher the number 
 *		the more information.  0 will provide no internal debugging.
 *
 * bbDebugLink	Used to get general status information about a single slave
 * bbDebugNode	node.
 *		To disable this feature, set one of them to -1.
 *
 *****************************************************************************/
int	bbDebug = 1;
int	bbDebugLink=-1;
int	bbDebugNode=-1;

#ifdef BB_SUPER_DEBUG
static int BBSetHistEvent(int link, struct dpvtBitBusHead *pXact, int State);
#endif

/******************************************************************************
 *
 * This structure contains a list of the outside-callable functions.
 *
 ******************************************************************************/
struct drvBitBusEt drvBitBus = {
  3,
  reportBB,
  initBB,
  qBBReq
};

STATIC BitbusLinkStruct	*pBBLink[BB_NUM_LINKS];	/* NULL if link not config'd */
/***************************************************************************
 *
 * User-callable configuration function.
 *
 * This function is used to configure the type of link as well as to 
 * initialize it for operation.
 *
 *****************************************************************************/
int BBConfig(unsigned long Link, 
	unsigned long LinkType,
	unsigned long BaseAddr,
	unsigned long IrqVector,
	unsigned long IrqLevel)
{
  void		*pVoid;
  int		j;
  static int	FirstTime = 1;
  char		nameTemp[30];
  epicsThreadId	taskId;
  long          status;
  unsigned int  priority;

  if (FirstTime)
  {
    rebootHookAdd((FUNCPTR)BBrebootFunc);
    FirstTime = 0;
  }
  if (Link >= BB_NUM_LINKS)
  {
    epicsPrintf("Error: Invalid link (%d) specified in BBConfig()\n", Link);
    return(-1);
  }
  if ((LinkType != BB_CONF_TYPE_XYCOM) && (LinkType != BB_CONF_TYPE_PEP))
  {
    epicsPrintf("Error: Invalid link type (%d) specified in BBConfig()\n", LinkType);
    return(-1);
  }
  if (pBBLink[Link] != NULL)
  {
    epicsPrintf("Error: BBConfig() Attempt to reconfigure link %d!\n", Link);
    return(-1);
  }

  pBBLink[Link] = (BitbusLinkStruct *) calloc(1,sizeof(BitbusLinkStruct));
  if(pBBLink[Link]==0)
  {
    epicsPrintf("Error: BBConfig() cannot calloc\n");
    return(-1);
  }

  pBBLink[Link]->LinkType = BB_CONF_HOSED;	/* Changed if all goes well */
  pBBLink[Link]->BaseAddr = BaseAddr;
  pBBLink[Link]->IrqVector = IrqVector;
  pBBLink[Link]->IrqLevel = IrqLevel;

#ifdef BB_SUPER_DEBUG
  /* Init the history FIFO to empty */
  pBBLink[Link]->History.sem = epicsMutexMustCreate();
  pBBLink[Link]->snoopSem = epicsMutexMustCreate();
  pBBLink[Link]->History.Next = 0;
  pBBLink[Link]->History.Num = 0;
#endif

  pBBLink[Link]->linkEventSem = epicsEventMustCreate(epicsEventEmpty);

  if (sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO , (char *)BaseAddr, (char **)&pVoid) != OK)

  {
    epicsPrintf("Error: BBConfig() can not translate requested A16 address(0x%8.8X\n", BaseAddr);
    pBBLink[Link] = NULL;
    return(-1);
  }

  /* 
   * Interface specific I/O mapping and registration 
   */

  switch (LinkType)
  {
  case BB_CONF_TYPE_XYCOM:
    pBBLink[Link]->l.XycomLink.rxInt = epicsEventMustCreate(epicsEventEmpty);
    pBBLink[Link]->l.XycomLink.bbRegs = (XycomBBRegsStruct *)pVoid;
    if (devRegisterAddress("Xycom Bitbus", atVMEA16,
    BaseAddr, sizeof(XycomBBRegsStruct), NULL) != 0)
    {
      epicsPrintf("Error: BBConfig() can not register address %p\n", pVoid);
      pBBLink[Link] = NULL;
      return(-1); /* BUG */
    }
    break;
  case BB_CONF_TYPE_PEP:
    pBBLink[Link]->l.PepLink.rxInt = epicsEventMustCreate(epicsEventEmpty);
    pBBLink[Link]->l.PepLink.bbRegs = (PepBBRegsStruct *)pVoid;
    if (devRegisterAddress("PEP Bitbus", atVMEA16,
    BaseAddr, sizeof(PepBBRegsStruct), NULL) != 0)
    {
      epicsPrintf("Error: BBConfig() can not register address %8.8X\n", pVoid);
      pBBLink[Link] = NULL;
      return(-1); /* BUG */
    }
    break;
  }

  /*
   * Common data structure initialization.
   */

  /* Init the prioritized queue lists */
  for (j=0; j<BB_NUM_PRIO; j++)
  {
    pBBLink[Link]->queue[j].head = NULL;
    pBBLink[Link]->queue[j].tail = NULL;
    pBBLink[Link]->queue[j].sem = epicsMutexMustCreate();
    pBBLink[Link]->queue[j].elements = 0;
  }

  /* Init the busy message list */
  pBBLink[Link]->busyList.sem = epicsMutexMustCreate();
  pBBLink[Link]->busyList.head = NULL;
  pBBLink[Link]->busyList.tail = NULL;
  pBBLink[Link]->busyList.elements = 0;

  for (j=0; j<BB_APERLINK; j++)
  { /* Assume all nodes are IDLE with no delay imposed */
    pBBLink[Link]->deviceStatus[j] = BB_IDLE;
    pBBLink[Link]->syntheticDelay[j] = 0;
  }
  pBBLink[Link]->DelayCount = 0;

  pBBLink[Link]->watchDogId = wdCreate();
  pBBLink[Link]->watchDogSem = epicsEventMustCreate(epicsEventEmpty);

  /* Clear the link abort status */
  pBBLink[Link]->abortFlag = 0;
  pBBLink[Link]->txAbortAck = 0;
  pBBLink[Link]->rxAbortAck = 0;

  /* 
   * Interface type specific initialization of management tasks.
   */

  if (LinkType == BB_CONF_TYPE_XYCOM)
  {
    /* BOARD SPECIFIC INIT CODE for XYCOM */
    if (xvmeReset(Link) != 0)
    {
      pBBLink[Link] = NULL;
      return(-1);
    }
  
    pBBLink[Link]->LinkType = LinkType;
  
    /* attach the interrupt handler routines */
    status = devConnectInterrupt(intVME,IrqVector+1,xvmeIrqRcmd,(void *)(int)Link);
    if(status) {
        errPrintf(status,0,0,"BBConfig link %lu\n",Link);
        return(-1);
    }
    status = devConnectInterrupt(intVME,IrqVector+3,
        xvmeIrqRdav,(void *)(int)Link);
    if(status) {
        errPrintf(status,0,0,"BBConfig link %lu\n",Link);
        return(-1);
    }
  
    /* Start a task to manage the TX link */
    sprintf(nameTemp, "bbTx%d-xy", (int)Link);
    epicsThreadHighestPriorityLevelBelow(epicsThreadPriorityHigh,&priority);
    taskId = epicsThreadCreate(nameTemp,
        priority, epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)xvmeTxTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start TX link task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
    /* Start a task to manage the RX link */
    sprintf(nameTemp, "bbRx%d-xy", (int)Link);
    taskId = epicsThreadCreate(nameTemp,
        epicsThreadPriorityHigh,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)xvmeRxTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start RX link task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
    /* Start a task to keep an eye on the busy list */
    sprintf(nameTemp, "bbwd%d-xy", (int)Link);
    epicsThreadLowestPriorityLevelAbove(epicsThreadPriorityHigh,&priority);
    taskId = epicsThreadCreate(nameTemp,
        priority, epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)xvmeWdTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start watchdog task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
  }
  else if (LinkType == BB_CONF_TYPE_PEP)
  {
    /* PEP specific init code */
    if (pepReset(Link) != 0)
    {
      pBBLink[Link] = NULL;
      return(-1);
    }
  
    pBBLink[Link]->LinkType = LinkType;
  
    /* attach the interrupt handler routines */
    status = devConnectInterrupt(intVME,IrqVector,pepIrqRdav,(void *)(int)Link);
    if(status) {
        errPrintf(status,0,0,"BBConfig link %lu\n",Link);
        return(-1);
    }
  
    /* Start a task to manage the TX link */
    sprintf(nameTemp, "bbTx%d-pep", (int)Link);
    epicsThreadHighestPriorityLevelBelow(epicsThreadPriorityHigh,&priority);
    taskId = epicsThreadCreate(nameTemp,
        priority, epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)pepTxTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start TX link task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
    /* Start a task to manage the RX link */
    sprintf(nameTemp, "bbRx%d-pep", (int)Link);
    taskId = epicsThreadCreate(nameTemp,
        epicsThreadPriorityHigh,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)pepRxTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start RX link task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
    /* Start a task to keep an eye on the busy list */
    sprintf(nameTemp, "bbwd%d-pep", (int)Link);
    epicsThreadLowestPriorityLevelAbove(epicsThreadPriorityHigh,&priority);
    taskId = epicsThreadCreate(nameTemp,
        priority, epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)pepWdTask,(void *)Link);
    if(!taskId) {
      epicsPrintf("BBConfig(): failed to start watchdog task for link %d\n", Link);
    }
    taskwdInsert(taskId,NULL,NULL);
  }
  status = devEnableInterruptLevel(intVME,IrqLevel);
  if(status) {
      errPrintf(status,0,0,"BBConfig link %lu\n",Link);
      return(-1);
  }
  return(0);
}
/******************************************************************
 *
 * The EPICS init routine is not needed.
 *
 ******************************************************************/
STATIC long initBB(void)
{
  return(0);
}

/******************************************************************
 * FUNCTION:    reportBB()
 * PURPOSE :    Prints a message indicating the presence of each
 *              bitbus module found in system.
 * ARGS IN :    none
 * ARGS OUT:    none
 * GLOBALS:     checks pBBLink[i] for NULL
 ******************************************************************/

STATIC long reportBB(void)
{
  int	i;

  if (bbDebug>1)
    printf("Bitbus debugging flag is set to %d\n", bbDebug);

  for (i=0; i< BB_NUM_LINKS; i++)
  {
    if (pBBLink[i] != NULL)
    {
      printf("Bitbus link %d pBBLink %p",i,pBBLink[i]);
      if (pBBLink[i]->LinkType == BB_CONF_TYPE_XYCOM) {
        printf("  %p IV=(0x%2.2lX,0x%2.2lX) IL=%lu (XYCOM)\n",
          pBBLink[i]->l.XycomLink.bbRegs,
          pBBLink[i]->IrqVector+1,pBBLink[i]->IrqVector+3,
          pBBLink[i]->IrqLevel);
        printf("            IV(0x%2.2lX,0x%2.2lX) reserved but not used\n",
          pBBLink[i]->IrqVector,pBBLink[i]->IrqVector+2);
      } else if (pBBLink[i]->LinkType == BB_CONF_TYPE_PEP) {
        printf(" %p IV=0x%2.2lX IL=%lu (PEP)\n",
          pBBLink[i]->l.PepLink.bbRegs, pBBLink[i]->IrqVector,
          pBBLink[i]->IrqLevel);
        printf("            IV=0x%2.2lX reserved but not used\n",
          pBBLink[i]->IrqVector+1);
      }
    }
  }
  return(OK);
}

int BBSnoop(int link,int node,int seconds)
{
    if(link<0 || link>=BB_NUM_LINKS || !pBBLink[link]) {
        printf("Illegal link\n");
        return(-1);
    }
    pBBLink[link]->nodeToSnoop = (char)node;
    pBBLink[link]->snoopEndTicks = tickGet() + seconds*sysClkRateGet();
    return(0);
}

static void snoopMessage(char *message,int link, struct bitBusMsg *pMsg)
{
  BitbusLinkStruct *pBitbusLinkStruct = pBBLink[link];
  static char buffer[200];
  char *pbuffer;
  int len;
  int i;

  if(pMsg->node != pBitbusLinkStruct->nodeToSnoop) return;
  if((unsigned long)tickGet() >= pBitbusLinkStruct->snoopEndTicks) return;
  epicsMutexMustLock(pBitbusLinkStruct->snoopSem);
  epicsThreadSleep(.05);
  pbuffer = buffer;
  len=sprintf(pbuffer,"bitBus %s: link=%4.4x node=%2.2x route=%2.2x tasks=%2.2x cmd=%2.2x length=%2.2x\n",
      message, (unsigned int)link, (unsigned int)pMsg->node,
      (unsigned int)pMsg->route, (unsigned int)pMsg->tasks,
      (unsigned int)pMsg->cmd,(unsigned int)pMsg->length);
  pbuffer += len;

  if (pMsg->length > BB_MSG_HEADER_SIZE) {
      len = sprintf(pbuffer,"    data:");
      pbuffer += len;
      for(i = 0; i < pMsg->length - BB_MSG_HEADER_SIZE; i++) {
	  len = sprintf(pbuffer," %2.2x",pMsg->data[i]);
	  pbuffer += len;
      }
      len = sprintf(pbuffer," |");
      pbuffer += len;
      for(i = 0; i < pMsg->length - BB_MSG_HEADER_SIZE; i++) {
	  if(iscntrl(pMsg->data[i]) || !isprint(pMsg->data[i]))
              len = sprintf(pbuffer,".");
	  else 
              len = sprintf(pbuffer,"%c",pMsg->data[i]);
	  pbuffer += len;
      }
      len = sprintf(pbuffer,"|\n");
  }
  fprintf(stderr,"%s",buffer);
  fflush(stderr);
  epicsMutexUnlock(pBitbusLinkStruct->snoopSem);
}
/******************************************************************
 * FUNCTION:    pepReset()
 * PURPOSE :    Performs firmware reset of PB-BIT module corresponding
 *              to link. Attempts to empty any data sitting in receive
 *              FIFO. Interrupts are disabled and Rx task is unblocked.
 * ARGS IN :    xvmeRegs ptr to register structure of PB-BIT module
 *              link link number serviced by PB-BIT module
 * ARGS OUT:    none
 * GLOBALS:     twiddles board
 ******************************************************************/
STATIC int
pepReset(int link)
{
  char  trash;
  int   j;
  int   lockKey;
  int	probeValue;

  if (bbDebug)
    printf("pepReset(%d): Resetting pep module\n", link);

  probeValue = 0;
  if (vxMemProbe(&(pBBLink[link]->l.PepLink.bbRegs->int_vec), WRITE, 1, &probeValue) < OK)
  { /* no BB board present here */
    epicsPrintf("ERROR: PEP Bitbus link %d not present at %p\n", link, pBBLink[link]->l.PepLink.bbRegs);
    return(-1);
  }

  /* Write firmware reset package (2 bytes) to board */
  pBBLink[link]->l.PepLink.bbRegs->data = 0x83;
  pBBLink[link]->l.PepLink.bbRegs->stat_ctl = 0x01;

  epicsThreadSleep(.3); /* give the 80152 time to self check */

  if ((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & 0x10) != 0x0) 
  {
    if (bbDebug)
      printf("pepReset(%d): PB-BIT firmware reset failed!\n", link);
    return(ERROR);
  }

  j = 1026;		/* 1K deep receive fifo */
			/* flush receive fifo if junk in it */
  while ((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & PEP_BB_RFNE) && --j)
    trash = pBBLink[link]->l.PepLink.bbRegs->data;

  if (!j) 
  {
    if (bbDebug)
      printf("pepReset(%d): receive fifo will not clear after reset!\n", link);
    return(ERROR);
  }

  /* Disable interrupts */
  lockKey = intLock();
  pBBLink[link]->l.PepLink.bbRegs->int_vec = 0;
  intUnlock(lockKey);

  /* Tell pepRxTask to Re-enable ints */
printf("pepReset calling epicsEventSignal link %d\n",link);
  epicsEventSignal(pBBLink[link]->l.PepLink.rxInt);  

  return(OK);
}
/****************************************************************************
 *
 * Reset an xvme-402 BitBus card by cycling the reset bit in the fifo
 * status register.
 *
 ****************************************************************************/
STATIC int xvmeReset(int link)
{
  char	trash;
  int	j;
  int	lockKey;
  unsigned char	probeValue;


  if (bbDebug)
    printf("xvmeReset(%d): Resetting xvme module\n", link);

  probeValue = XVME_RESET;
  if (vxMemProbe(&(pBBLink[link]->l.XycomLink.bbRegs->fifo_stat), WRITE, 1, &probeValue) < OK)
  { /* no BB board present here */
    epicsPrintf("ERROR: Xycom Bitbus link %d not present at %p\n", link, pBBLink[link]->l.XycomLink.bbRegs);
    return(-1);
  }

#ifdef BB_SUPER_DEBUG
  BBSetHistEvent(link, NULL, XACT_HIST_STATE_RESET);
#endif

  epicsThreadSleep(.05);
  pBBLink[link]->l.XycomLink.bbRegs->fifo_stat = 0;	/* clear reset pulse */
  epicsThreadSleep(.5); /* give the 8044 time to self check */

  j = 100;				/* give up after this */
  while ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_RCMD) && --j)
    trash = pBBLink[link]->l.XycomLink.bbRegs->cmnd;		/* flush command buffer if junk in it */

  if (!j)
  {
    if (bbDebug)
      printf("xvmeReset(%d): Command buffer will not clear after reset!\n", link);
    return(ERROR);
  }

  j = 100;
  while ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_RFNE) && --j)
    trash = pBBLink[link]->l.XycomLink.bbRegs->data;	/* flush data buffer if junk in it */

  if (!j)
  {
    if (bbDebug)
      printf("xvmeReset(%d): Data buffer will not clear after reset!\n", link);
    return(ERROR);
  }

  if ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_FSVALID) != XVME_FSIDLE)
  {
    if (bbDebug)
      printf("xvmeReset(%d): XVME board not returning to idle status after reset!\n", link);
    return(ERROR);
  }

  /* set the interrupt vector */
  lockKey = intLock();
  pBBLink[link]->l.XycomLink.bbRegs->stat_ctl = 0;          /* disable all interupts */
  pBBLink[link]->l.XycomLink.bbRegs->int_vec = pBBLink[link]->IrqVector;
  intUnlock(lockKey);

  epicsEventSignal(pBBLink[link]->l.XycomLink.rxInt);	/* Tell xvmeRxTask to Re-enable interrupts */

  return(OK);
}
/****************************************************************************
 *
 * This function is used to add a node to the HEAD of a list.
 *
 ****************************************************************************/
static int
listAddHead(struct bbList *plist, struct dpvtBitBusHead *pnode)
{
  pnode->prev = NULL;
  pnode->next = plist->head;

  if (plist->head != NULL)
    plist->head->prev = pnode;

  if (plist->tail == NULL)
    plist->tail = pnode;

  plist->head = pnode;

  plist->elements++;

  return(0);
}
/******************************************************************************
 *
 * This function is used to add a node to the TAIL of a list.
 *
 ******************************************************************************/
static int
listAddTail(struct bbList *plist, struct dpvtBitBusHead *pnode)
{
  pnode->next = NULL;		/* No next node if this is the TAIL */
  pnode->prev = plist->tail;	/* previous node is the 'old' TAIL node */

  if (plist->tail != NULL)
    plist->tail->next = pnode;	/* link the 'old' tail to the 'new' tail node */

  if (plist->head == NULL)
    plist->head = pnode;

  plist->tail = pnode;		/* this is the 'new' tail node */

  plist->elements++;

  return(0);
}
/***************************************************************************
 *
 * This function is used to delete a node from a list.
 *
 ***************************************************************************/
static int
listDel(struct bbList *plist, struct dpvtBitBusHead *pnode)
{
  if (pnode->next != NULL)
    pnode->next->prev = pnode->prev;

  if (pnode->prev != NULL)
    pnode->prev->next = pnode->next;

  if (plist->head == pnode)
    plist->head = pnode->next;

  if (plist->tail == pnode)
    plist->tail = pnode->prev;

  pnode->next = pnode->prev = 0;
  plist->elements--;
  if(plist->elements<=0) {
    if(plist->elements<0 ||(plist->head!=0 || plist->tail!=0)) {
      printf("drvBitBus(listDel) plist %p elements %d head %p tail %p\n",
		plist, plist->elements, plist->head, plist->tail);
      epicsThreadSuspendSelf();
    }
  }
  return(0);
}

static int listIsOnList(int link, struct dpvtBitBusHead *pnode)
{
  BitbusLinkStruct *pBitbusLinkStruct = pBBLink[link];
  int ind;
  struct bbList *pbblist;
  struct dpvtBitBusHead *pnext;

printf("listIsOnList link %d pnode %p\n",link,pnode);
  for(ind=0; ind<BB_NUM_PRIO+1; ind++) {
    if(ind<BB_NUM_PRIO) pbblist = &pBitbusLinkStruct->queue[ind];
    else pbblist = &pBitbusLinkStruct->busyList;
    pnext = pbblist->head;
    while(pnext) {
      if(pnext==pnode)  {
        printf("link %d node %p already on a list\n",link,pnode);
        return(-1);
      }
      pnext = pnext->next;
    }
  }
  return(0);
}
/****************************************************************************
 * 
 * xycom IRQ handler for receiver data bytes.
 *
 ****************************************************************************/
static void
xvmeIrqRdav(int link)
{
  if (bbDebug > 30)
    logMsg("bitbus rx IRQ on link %d\n", link,0,0,0,0,0);
  epicsEventSignal(pBBLink[link]->l.XycomLink.rxInt);	/* deliver the groceries */
}

/****************************************************************************
 *
 * xycom IRQ handler invoked when the BitBus controller has completed the 
 * transfer of a RECEIVED message.  
 *
 ****************************************************************************/
static void
xvmeIrqRcmd(int link)
{
  if (bbDebug > 30)
    logMsg("bitbus rcmd IRQ on link %d\n", link,0,0,0,0,0);
  epicsEventSignal(pBBLink[link]->l.XycomLink.rxInt);      /* deliver the groceries */
}

/******************************************************************
 * FUNCTION:    pepIrqRdav()
 * PURPOSE :    Invoked when PB-BIT module has received a complete
 *              bitbus message (ie. not on a byte-by-byte basis).
 * ARGS IN :    link link number upon which message has arrived
 * ARGS OUT:    none
 * GLOBALS:     unblocks Rx task (if it is not already running)
 ******************************************************************/
STATIC void
pepIrqRdav(int link)
{
  pBBLink[link]->l.PepLink.bbRegs->int_vec = 0;	/* disable IRQs */

  if (bbDebug > 30)
    logMsg("PEP bitbus rx IRQ on link %d\n", link,0,0,0,0,0);

  epicsEventSignal(pBBLink[link]->l.PepLink.rxInt);       /* unblock it */
}
/******************************************************************
 * FUNCTION:    pepTmoHandler()
 * PURPOSE :    Invoked whenever a watchdog timer times out. Watchdogs
 *              are running whenever the busyList has any elements on
 *              it. The idea here is that the watchdog handler scans
 *              through the busyList, looking for old requests that have
 *              not been replied to in too long a time. If there are
 *              any old ones around, they are removed from the list,
 *              and the associated node is reset and marked offline.
 *
 * ARGS IN :    link
 * ARGS OUT:    none
 * GLOBALS:     none
 ******************************************************************/
STATIC int
pepTmoHandler(int link)
{
  if (bbDebug > 25)
    logMsg("pepTmoHandler(%d): Watch dog interrupt\n", link,0,0,0,0,0);

  epicsEventSignal(pBBLink[link]->watchDogSem); /* unblock Wd task */
  return(0);
}

/******************************************************************************
 *
 * Watchdogs are running when ever the busy list has any elements in it.
 * The idea here is that the watchdog handler scans thru the busy list,
 * looking for old requests that have not been replied to in too long
 * a time.  If there are any old ones around, they are removed from the
 * list and marked as un-replied to.
 *
 ******************************************************************************/
static int
xvmeTmoHandler(int link)
{
  if (bbDebug > 25)
    logMsg("xvmeTmoHandler(%d): Watch dog interrupt\n", link,0,0,0,0,0);

  epicsEventSignal(pBBLink[link]->watchDogSem);
  return(0);
}

/****************************************************************************
 *
 * Given a link number, make sure it is valid.
 *
 ****************************************************************************/
static int
checkLink(int link)
{
  if ((link<0) || (link>BB_NUM_LINKS))
    return(ERROR);		/* link number out of range */

  if (pBBLink[link] == NULL)
    return(ERROR);		/* link number has no card installed */

  return(OK);
}

/******************************************************************************
 *
 * This function is called by user programs to queue an I/O transaction request
 * for the BB driver.  It is the only user-callable function provided in this
 * driver.
 *
 ******************************************************************************/
STATIC long qBBReq(struct  dpvtBitBusHead *pdpvt, int prio)
{
  static linkErrFlags[BB_NUM_LINKS];    /* Supposedly init'd to zero */

  if ((prio < 0) || (prio >= BB_NUM_PRIO))
  {
    epicsPrintf("invalid priority requested in call to qbbreq(%p, %d)\n",
        pdpvt, prio);
    return(ERROR);
  }
  if (checkLink(pdpvt->link) == ERROR)
  {
    if (pdpvt->link >= BB_NUM_LINKS) {
        epicsPrintf("qbbreq(%p, %d) %d\n", pdpvt, prio, pdpvt->link);
    } else if (linkErrFlags[pdpvt->link] == 0){/* Anti-message swamping check */
      linkErrFlags[pdpvt->link] = 1;
        epicsPrintf("qbbreq(%p, %d) %d... card not presen\n",
            pdpvt, prio, pdpvt->link);
    }
    return(ERROR);
  }

#ifdef BB_SUPER_DEBUG
  BBSetHistEvent(pdpvt->link, pdpvt, XACT_HIST_STATE_QUEUE);
#endif

  if (bbDebug>5)
    printf("qbbreq(%p, %d): transaction queued\n", pdpvt, prio);
  if (bbDebug>6)
    drvBitBusDumpMsg(&(pdpvt->txMsg));

  if(pdpvt->next || pdpvt->prev) {
    if(listIsOnList(pdpvt->link,pdpvt)) return(ERROR); 
  }
  snoopMessage("qBBReq",pdpvt->link,&pdpvt->txMsg);

  epicsMutexMustLock(pBBLink[pdpvt->link]->queue[prio].sem);
  /* Add to the end of the queue of waiting transactions */
  listAddTail(&(pBBLink[pdpvt->link]->queue[prio]), pdpvt);

  epicsMutexUnlock(pBBLink[pdpvt->link]->queue[prio].sem);

  epicsEventSignal(pBBLink[pdpvt->link]->linkEventSem);

  return(OK);
}

/******************************************************************
 *
 * Reset hook function... takes down ALL the bitbus links and leaves
 * them down.
 *
 ******************************************************************/
STATIC void BBrebootFunc(void)
{
  int i;

  for (i=0; i<BB_NUM_LINKS; i++)
  {
    if (checkLink(i) == OK)
      bbKill(i);
  }
  epicsThreadSleep(1.0);
  return;
}

/******************************************************************************
 *
 * A user callable link resetter.  This sets a flag and releases the dog
 * task to reset the link.
 *
 ******************************************************************************/
int
BBReset(int link)
{
  if (checkLink(link) != ERROR)
  {
    pBBLink[link]->nukeEm = 1;
    epicsEventSignal(pBBLink[link]->watchDogSem);
  }
  else
    printf("Link %d not installed.\n", link);

  return(0);
}

/*******************************************************************
 *
 * Same as BBReset() but it takes the link down and leaves it down.
 *
 *******************************************************************/
int bbKill(int link)
{
  if (checkLink(link) != ERROR)
  {
    pBBLink[link]->nukeEm = 2;
    epicsEventSignal(pBBLink[link]->watchDogSem);
  }
  else
    printf("Link %d not installed.\n", link);

  return(0);
}

/******************************************************************************
 *
 * This gets called by the transmit task if it gets stuck waiting to send a 
 * byte to the transmit fifo.
 *
 ******************************************************************************/
STATIC int txStuck(int link)
{
  if (bbDebug)
    printf("bitbus transmitter task stuck, resetting link %d\n", link);

  BBReset(link);
  while (pBBLink[link]->abortFlag == 0) epicsThreadSleep(.05);

  return(OK);
}

/******************************************************************************
 *
 * A user callable function that may be used to reset the specified slave
 * node.
 *
 * This routine is NOT reenterant but contains a monitor so that it always
 * operates properly.
 *
 ******************************************************************************/
int BBResetBug(int link, int node)
{
  static int            FirstTime = 1;
  static epicsMutexId   monLock;
  static epicsEventId   syncSem;        /* Semaphore to use for completion */
  struct dpvtBitBusHead resetNode;      /* actual message to send out */
  unsigned char         resetNodeData;  /* 1-byte data field for RAC_RESET */
  if (checkLink(link) == ERROR)
  {
    printf("Link %d not installed.\n", link);
  }
  if (FirstTime)
  {
    monLock = epicsMutexMustCreate();
    syncSem = epicsEventMustCreate(epicsEventEmpty);
    FirstTime = 0;
  }

  epicsMutexMustLock(monLock);

  callbackSetCallback(0,&resetNode.callback);
  resetNode.psyncSem = &syncSem;/* do a semGive on this SEM when done sending */
  resetNode.next = 0;
  resetNode.prev = 0;
  resetNode.link = link;        /* which bitbus link to send message out on */
  resetNode.rxMaxLen = BB_MSG_HEADER_SIZE;       /* Chop off the response...*/
  resetNode.ageLimit = sysClkRateGet()*2;

  resetNode.txMsg.length = 8;
  resetNode.txMsg.route = BB_STANDARD_TX_ROUTE;
  resetNode.txMsg.node = node;
  resetNode.txMsg.tasks = 0x0;
  resetNode.txMsg.cmd = RAC_RESET_SLAVE;
  resetNode.txMsg.data = &resetNodeData;

  qBBReq(&resetNode, 0);
  epicsEventMustWait(syncSem);
  epicsMutexUnlock(monLock);

  return(0);
}
 /***************************************************************************
 *
 * This function is started as a task during driver init time.  It's purpose
 * is to read messages that are received from the bitbus link.  After a message
 * is received, the soliciting message's completion routines are started by
 * the use of semaphores and/or a callbackRequest().
 *
 ***************************************************************************/
#define BBRX_HEAD	1
#define	BBRX_DATA	2
#define BBRX_RCMD	3
#define	BBRX_IGN	4
void xvmeRxTask(int link)
{
  int           rxState;       /* current state of the receiver */

  unsigned char rxHead[5];      /* room for header of current rx msg */
  unsigned char *rxMsg;         /* where to place next byte (after header) */
  int           rxTCount;       /* byte counter for data in rxHead */
  unsigned char	ch;
  struct dpvtBitBusHead *rxDpvtHead;	/* for message currently receiving */

  int		lockKey;		/* used for intLock calls */

  rxMsg = (unsigned char *) NULL;
  rxState = BBRX_HEAD;
  rxTCount = 0;
  rxDpvtHead = (struct dpvtBitBusHead *) NULL;

  while (1)
  {
    /* Wait for RX interrupts, but only if no chars are ready */
    if ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_RFNE) == 0)
    {
      /* Enable interrupts and check again because xycom blew it */
      lockKey = intLock();
      pBBLink[link]->l.XycomLink.bbRegs->stat_ctl = XVME_ENABLE_INT | XVME_RX_INT;
      intUnlock(lockKey);
      
      while (((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_RFNE) == 0)
      && (pBBLink[link]->abortFlag == 0))
      { 
        /* Re-enable ints here each time in case board got reset */
	lockKey = intLock();
        pBBLink[link]->l.XycomLink.bbRegs->stat_ctl = XVME_ENABLE_INT | XVME_RX_INT;
	intUnlock(lockKey);
  
	/* Wait for groceries */
        epicsEventMustWait(pBBLink[link]->l.XycomLink.rxInt);
      }
      /* Disable RX Interrupts (prevents unnecessary context switching) */
      lockKey = intLock();
      pBBLink[link]->l.XycomLink.bbRegs->stat_ctl = 0;
      intUnlock(lockKey);
    }
    if (pBBLink[link]->abortFlag == 0)
    {
    /* check to see if we got a data byte or a command byte */
    if ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_RCMD) == XVME_RCMD)
      rxState = BBRX_RCMD;

    switch (rxState) {
    case BBRX_HEAD:	/* getting the head of a new message */
      rxHead[rxTCount] = pBBLink[link]->l.XycomLink.bbRegs->data;
      if (bbDebug>21)
        printf("xvmeRxTask(%d): >%2.2X< (Header)\n", link, rxHead[rxTCount]);

      if (++rxTCount == 5)
      { /* find the message this is a reply to */
	rxTCount += 2;		/* adjust for the link field space */

	/* Lock the busy list */
	epicsMutexMustLock(pBBLink[link]->busyList.sem);

        rxDpvtHead = pBBLink[link]->busyList.head;
        while (rxDpvtHead != NULL)
        {
          if (bbDebug>19)
            printf("xvmeRxTask(%d): checking reply against %p\n",
            link, rxDpvtHead);

          /* see if node's match */
          if (rxDpvtHead->txMsg.node == rxHead[2])
          { /* see if the tasks match */
            if (rxDpvtHead->txMsg.tasks == rxHead[3])
            { /* They match, finish putting response into the rxMsg buffer */
              if (bbDebug>4)
                printf("xvmeRxTask(%d): reply to %p\n", link, rxDpvtHead);
  
              /* Delete the node from the list */
	      listDel(&(pBBLink[link]->busyList), rxDpvtHead);

	      /* If busy list is empty, stop the dog */
              if (pBBLink[link]->busyList.head == NULL) 
                wdCancel(pBBLink[link]->watchDogId);
          
              if (rxHead[4] == 0x91)
	      { /* something bad happened... inject a delay to the */
		/* requested timeout duration. */

                if ((link == bbErrorTrigLink) &&
                    ((bbErrorTrigNode < 0) || (bbErrorTrigNode == rxHead[2])))
                    pulseSysMon();

                /* keep track of timeouts per node */
                if(rxHead[2] < BB_NUM_NODES) {
                    bbNode91[link][rxHead[2]]++;
                }

		if (bbDebug)
		  printf("xvmeRxTask(%d): 0x91 node %d, start synthetic delay\n"
                      , link, rxHead[2]);
		(pBBLink[link]->syntheticDelay[rxDpvtHead->txMsg.node])
                    = rxDpvtHead->retire;
		pBBLink[link]->DelayCount++;
              }
              else
              { /* decrement the number of outstanding messages to the node */
                (pBBLink[link]->deviceStatus[rxDpvtHead->txMsg.node])--;
	      }

	      /* Wake up Link Task in case was waiting on "this" node */
              epicsEventSignal(pBBLink[link]->linkEventSem); 
              break;	/* get out of the while() */
            }
          }
          rxDpvtHead = rxDpvtHead->next; /* Keep looking */
        }

	epicsMutexUnlock(pBBLink[link]->busyList.sem);

        if (rxDpvtHead == NULL) {
	    if (bbDebug)
	    {
              printf("xvmeRxTask(%d): msg from node %d unsolicited:",
              link, rxHead[2]);
	    }
            rxState = BBRX_IGN;
            break;
        }
        rxDpvtHead->rxMsg.length = rxHead[0];
        rxDpvtHead->rxMsg.route = rxHead[1];
        rxDpvtHead->rxMsg.node = rxHead[2];
        rxDpvtHead->rxMsg.tasks = rxHead[3];
	rxDpvtHead->rxMsg.cmd = rxHead[4];
        rxMsg = rxDpvtHead->rxMsg.data;
        rxDpvtHead->status = BB_OK;	/* OK, unless BB_LENGTH */
        rxState = BBRX_DATA;		/* finish reading till RCMD */
      }
      break;

    case BBRX_DATA:	/* finish reading data portion of message */
      ch = pBBLink[link]->l.XycomLink.bbRegs->data;
      if (rxTCount >= rxDpvtHead->rxMaxLen)
      {
        rxState = BBRX_IGN;	/* toss the rest of the data */
        rxDpvtHead->status = BB_LENGTH; /* set driver status */
	if (bbDebug>22)
	  printf("xvmeRxTask(%d): %2.2X (Ignored)\n", link, ch);
      }
      else
      {
        *rxMsg = ch;
        if (bbDebug>22)
          printf("xvmeRxTask(%d): %2.2X (Data)\n", link, ch);
        rxMsg++;
        rxTCount++;
      }
      break;

    case BBRX_IGN:
      ch = pBBLink[link]->l.XycomLink.bbRegs->data;
      if (bbDebug>22)
        printf("xvmeRxTask(%d): %2.2X (Ignored)\n", link, ch);
      break;

    case BBRX_RCMD:
      if (rxDpvtHead == NULL)
      {
        ch = pBBLink[link]->l.XycomLink.bbRegs->cmnd;
	if (bbDebug)
          printf("xvmeRxTask(%d): got unexpected XVME_RCMD\n", link);
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, NULL, XACT_HIST_STATE_RX_URCMD);
#endif
      }
      else
      {
        rxDpvtHead->status = BB_OK;	/* XXX -- ??? */
        rxDpvtHead->rxCmd = pBBLink[link]->l.XycomLink.bbRegs->cmnd;

        if (bbDebug>24)
          printf("xvmeRxTask(%d):RX command byte = %2.2X\n",
              link, rxDpvtHead->rxCmd);
    
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, rxDpvtHead, XACT_HIST_STATE_RX);
#endif
        snoopMessage("xvmeRxTask",link,&rxDpvtHead->rxMsg);
        if (rxDpvtHead->callback.callback != NULL)
        {
          if (bbDebug>8)
            printf("xvmeRxTask(%d): invoking the callbackRequest\n", link);
          callbackRequest(&rxDpvtHead->callback); /* schedule completion processing */
        } else if (rxDpvtHead->psyncSem != NULL) {
            epicsEventSignal(*(rxDpvtHead->psyncSem));
	}
      }
      /* Reset the state of the RxTask to expect a new message */
      rxMsg = (unsigned char *) NULL;
      rxState = BBRX_HEAD;
      rxTCount = 0;
      rxDpvtHead = (struct dpvtBitBusHead *) NULL;

      break;
    }
  }
  else
  { /* Link abort state is active reset receiver link status now */
    if (rxDpvtHead != NULL)
    { /* This xact is not on the busy list, put it back on */
      epicsMutexMustLock(pBBLink[link]->busyList.sem);
      (pBBLink[link]->deviceStatus[rxDpvtHead->txMsg.node])++;
      listAddTail(&(pBBLink[link]->busyList), rxDpvtHead);
      epicsMutexUnlock(pBBLink[link]->busyList.sem);
    }

#ifdef BB_SUPER_DEBUG
    BBSetHistEvent(link, rxDpvtHead, XACT_HIST_STATE_RX_ABORT);
#endif
    rxMsg = (unsigned char *) NULL;
    rxState = BBRX_HEAD;
    rxTCount = 0;
    rxDpvtHead = (struct dpvtBitBusHead *) NULL;

    /* Tell the watch dog I am ready for the reset (reset in the dog task) */
    pBBLink[link]->rxAbortAck = 1;

    if (bbDebug)
      printf("xvmeRxTask(%d): resetting due to abort status\n", link);

    /* wait for link state to become active again */
    while (pBBLink[link]->abortFlag != 0) epicsThreadSleep(RESET_POLL_TIME);

    if (bbDebug)
      printf("xvmeRxTask(%d): restarting after abort\n", link);
  }
  }
}

/******************************************************************************
 *
 * BUG -- this does not do the nuke=2 reset and kill link operations like
 * the PEP does.
 *
 ******************************************************************************/
void xvmeWdTask(int link)
{
  struct dpvtBitBusHead *pnode;
  struct dpvtBitBusHead *npnode;
  unsigned      long    now;


  pBBLink[link]->nukeEm = 0;		/* Make sure the nuke status is clear */
  while(1)
  {
    epicsEventMustWait(pBBLink[link]->watchDogSem);
    now = tickGet();		/* what time is it? */

    if (pBBLink[link]->nukeEm != 0)
      printf("Bitbus manual reset being issued on link %d\n", link);

    if (bbDebug>4)
      printf("xvmeWdTask(%d): (Watchdog) checking busy list\n", link);

    pBBLink[link]->rxAbortAck = 0;      /* In case we need to use them */
    pBBLink[link]->txAbortAck = 0;

    if (pBBLink[link]->nukeEm != 0)
    { /* set abort status and wait for the abort acks */
      pBBLink[link]->abortFlag = 1;

      /* wake up the Tx task so it can observe the abort status */
      epicsEventSignal(pBBLink[link]->linkEventSem);

      /* wake up the Rx task so it can observe the abort ststus */
      epicsEventSignal(pBBLink[link]->l.XycomLink.rxInt);

      /* sleep until abort ack from Tx & Rx tasks */
      while ((pBBLink[link]->rxAbortAck == 0) && (pBBLink[link]->txAbortAck == 0))
	epicsThreadSleep(RESET_POLL_TIME);

#ifdef BB_SUPER_DEBUG
      if (pBBLink[link]->nukeEm < 2)
      { /* Do a history dump since this is not supposed to happen */
        BBHistDump(link);
      }
#endif
    }
    /*
     * Run thru entire busy list to see if there are any transactions
     * that have been waiting on a response for too long a period.
     */
    epicsMutexMustLock(pBBLink[link]->busyList.sem);
    pnode = pBBLink[link]->busyList.head;

    while (pnode != NULL)
    {
      npnode = pnode->next;   /* remember where we were in the list */

      if ((pBBLink[link]->nukeEm != 0) || (pnode->retire <= now))
      {
	/* Get rid of the request and set error status etc... */
	listDel(&(pBBLink[link]->busyList), pnode);

	if (bbDebug)
	{
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, pnode, XACT_HIST_STATE_WD_TIMEOUT);
#endif
	  printf("xvmeWdTask(%d): TIMEOUT on bitbus message:\n", link);
	  drvBitBusDumpMsg(&pnode->txMsg);
	}

        if(pnode->txMsg.node < BB_NUM_NODES) {
            bbNodeTMO[link][pnode->txMsg.node]++;
        }


	/* BUG -- do this here or defer until RX gets a response? */
        (pBBLink[link]->deviceStatus[pnode->txMsg.node])--; /* fix device status */
	pnode->status = BB_TIMEOUT;


	/* Make the callbackRequest if one was spec'd */
        snoopMessage("xvmeWdTask",link,&pnode->txMsg);
	if(pnode->callback.callback != NULL)
	{
	  callbackRequest(&pnode->callback);     /* schedule completion processing */
	}
	else
	{
	  /* Release a completion lock if one was spec'd */
	  if (pnode->psyncSem != NULL)
	    epicsEventSignal(*(pnode->psyncSem));
	}
	epicsEventSignal(pBBLink[link]->linkEventSem);
      }
      pnode = npnode;
    }
    if (pBBLink[link]->busyList.head != NULL)
    { /* Restart the dog timer */
      int ticks;

      if (bbDebug>5)
	printf("xvmeWdTask(%d): restarting watch dog timer\n", link);

      ticks = pBBLink[link]->busyList.head->retire - now;
      if(ticks<=0) {
	  printf("xvmeWdTask wdStart ticks %d\n",ticks);
	  ticks=1;
      }
      wdStart(pBBLink[link]->watchDogId, ticks, xvmeTmoHandler, link);
    }
    epicsMutexUnlock(pBBLink[link]->busyList.sem);

    /* Finish the link reboot if necessary */
    if (pBBLink[link]->nukeEm != 0)
    {
      xvmeReset(link);

      /* clear the abort_flag */
      pBBLink[link]->abortFlag = 0;

      pBBLink[link]->nukeEm = 0;
    }
  }
}

/******************************************************************************
 *
 * This function is started as a task during driver init time.  It's purpose
 * is to keep the link busy.  It awaits user-calls to qBbReq() with new work
 * as well as wake-up calls from the watchdog timer.
 *
 * At the time this function is started as its own task, the linked list
 * structures will have been created and initialized.
 *
 ******************************************************************************/
int xvmeminstuck = 1000;
void xvmeTxTask(int link)
{
  struct dpvtBitBusHead	*pnode;
  int			prio;
  int			working;
  int			dogStart;
  int			stuck;

  int			txTCount;
  int			txCCount;
  unsigned char		*txMsg;
  unsigned	long	now;
  epicsMutexId		resetNodeSem;
  struct dpvtBitBusHead	resetNode;
  unsigned char		resetNodeData;	/* 1-byte data field for RAC_OFFLINE */

  if (bbDebug)
    printf("xvmeTxTask started for link %d\n", link);

  /* hand-craft a RAC_OFFLINE  message for use with RAC_RESET_SLAVE commands */
  /* NOTE that having only one copy is OK provided that this message is    */
  /*      sent immediately following the RAC_RESET_SLAVE message.          */

  resetNodeSem = epicsMutexMustCreate();

  callbackSetCallback(0,&resetNode.callback);
  resetNode.psyncSem = &resetNodeSem;
  resetNode.next = 0;
  resetNode.prev = 0;
  resetNode.link = link;
  resetNode.rxMaxLen = BB_MSG_HEADER_SIZE;	/* Chop off the response */
  resetNode.ageLimit = sysClkRateGet();

  resetNode.txMsg.length = 8;
  resetNode.txMsg.route = BB_STANDARD_TX_ROUTE;
  resetNode.txMsg.node = 0xff;
  resetNode.txMsg.tasks = 0x0;
  resetNode.txMsg.cmd = RAC_OFFLINE;
  resetNode.txMsg.data = &resetNodeData;

  while(1)
  {
    if (pBBLink[link]->abortFlag != 0)
    {
      pBBLink[link]->txAbortAck = 1;	/* let the dog know we are ready */
#ifdef BB_SUPER_DEBUG
      BBSetHistEvent(link, NULL, XACT_HIST_STATE_TX_ABORT);
#endif

      if (bbDebug)
        printf("xvmeTxTask(%d): resetting due to abort status\n", link);

      while (pBBLink[link]->abortFlag != 0)
        epicsThreadSleep(RESET_POLL_TIME);	/* wait for link to reset */

      if (bbDebug)
        printf("xvmeTxTask(%d): restarting after abort\n", link);
    }
    else
    {
      if (pBBLink[link]->DelayCount)
	epicsEventWaitWithTimeout(pBBLink[link]->linkEventSem, 2.0);
      else
	epicsEventMustWait(pBBLink[link]->linkEventSem);
    }

    if (bbDebug>5)
      printf("xvmeTxTask(%d): got an event\n", link);

    working = 1;
    while ((working != 0)
    && (pBBLink[link]->abortFlag == 0)
    && (pBBLink[link]->busyList.elements < XycomMaxOutstandMsgs))
    {
      working = 0;

      prio = BB_NUM_PRIO-1;
      while ((prio >= 0) && (pBBLink[link]->abortFlag == 0))
      {
        /* see if the queue has anything in it */
	epicsMutexMustLock(pBBLink[link]->queue[prio].sem);
  
        if ((pnode = pBBLink[link]->queue[prio].head) != NULL)
        {
	  now = tickGet();
          while (pBBLink[link]->deviceStatus[pnode->txMsg.node] == BB_BUSY)
          {
	    if ((pBBLink[link]->syntheticDelay[pnode->txMsg.node] != 0)
		&& (pBBLink[link]->syntheticDelay[pnode->txMsg.node] < now))
	    {
	      if (bbDebug)
	        printf("xvmeTxTask(%d): terminating synthetic idle on node %d\n"
                    , link, pnode->txMsg.node);
	      (pBBLink[link]->deviceStatus[pnode->txMsg.node])--;
	      pBBLink[link]->syntheticDelay[pnode->txMsg.node] = 0;
	      pBBLink[link]-> DelayCount--;
	    }
	    else
	    {
              if ((pnode = pnode->next) == NULL)
                break;
	    }
	  }
        }
        if (pnode != NULL)
        { /* have an xact to start processing */
          working = 1;			/* indicate work being done */

          /* delete the node from the inbound fifo queue */
          listDel(&(pBBLink[link]->queue[prio]), pnode);
  
	  epicsMutexUnlock(pBBLink[link]->queue[prio].sem);
      
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX);
#endif
          if (bbDebug>3)
            printf("xvmeTxTask(%d): got xact, pnode=%p\n", link, pnode);
      
          /* Send the message in polled mode */
  
          txTCount = pnode->txMsg.length - 2;
          txCCount = 0;
          txMsg = &(pnode->txMsg.length);
 
          while ((txCCount <= txTCount) && (pBBLink[link]->abortFlag == 0))
          {
	    stuck = 1000;
            while(
            ((pBBLink[link]->l.XycomLink.bbRegs->fifo_stat & XVME_TFNF) != XVME_TFNF)
            && (pBBLink[link]->abortFlag == 0) && --stuck) {
                volatile int x;
		if(stuck > 900) for(x=0;x<100;x++);
		else epicsThreadSleep(.05);
	    }
if(stuck<xvmeminstuck) xvmeminstuck = stuck;
  
	    if (!stuck)
	    {
#ifdef BB_SUPER_DEBUG
  	      BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_STUCK);
#endif
	      txStuck(link);	/* will end up setting abortFlag */
	    }
            else if (txCCount < txTCount) /* on last time, just wait, no data */
	    {
              pBBLink[link]->l.XycomLink.bbRegs->data = *txMsg;	/* send next byte */
              if (bbDebug>30)
                printf("xvmeTxTask(%d): outputting %2.2X\n", link, *txMsg);
  
	      /* On 5th byte, we are dun w/header, set start w/data buffer */
	      if (txCCount != 4)
	        txMsg++;
              else
	        txMsg = pnode->txMsg.data;	
	    }
	    txCCount++;
          }

          snoopMessage("xvmeTxTask",link,&pnode->txMsg);
	  if (pBBLink[link]->abortFlag == 0)
	  {

	    /* All data bytes have been sent, put on busy list and release */

	    /* Don't add to busy list if was a RAC_RESET_SLAVE */
	    if ((pnode->txMsg.cmd == RAC_RESET_SLAVE)
            && (pnode->txMsg.tasks == 0))
	    { /* Finish the transaction here if was a RAC_RESET_SLAVE */
#ifdef BB_SUPER_DEBUG
              BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_RESET);
#endif
              if (bbDebug)
                printf("xvmeTxTask(%d): RAC_RESET_SLAVE sent, resetting node %d\n"
                    , link, pnode->txMsg.node);
  
	      pnode->status = BB_OK;
  
    	      if (pnode->callback.callback != NULL)
    	      {
      	        if (bbDebug>4)
	          printf("xvmeTxTask(%d): invoking callbackRequest\n", link);
	      
      	        callbackRequest(&pnode->callback); /* schedule completion processing */
    	      }
  	      else
	      {
    	        /* If there is a semaphore for synchronous I/O, unlock it */
    	        if (pnode->psyncSem != NULL)
      	          epicsEventSignal(*(pnode->psyncSem));
  	      }

              /* Wait for last NODE_OFFLINE to finish (if still pending) */
              epicsMutexMustLock(resetNodeSem);
  
	      /* have to reset the master so it won't wait on a response */
	      resetNodeData = pnode->txMsg.node; /* mark the node number */
	      epicsMutexMustLock(pBBLink[link]->queue[BB_Q_HIGH].sem);
	      listAddHead(&(pBBLink[link]->queue[BB_Q_HIGH]), &resetNode);
	      epicsMutexUnlock(pBBLink[link]->queue[BB_Q_HIGH].sem);

	    }
	    else
	    {
	      /* Lock the busy list */
	      epicsMutexMustLock(pBBLink[link]->busyList.sem);
  
              /* set the retire time */
              pnode->retire = tickGet();
	      if (pnode->ageLimit)
	        pnode->retire += pnode->ageLimit;
	      else
		pnode->retire += BB_DEFAULT_RETIRE_TIME * sysClkRateGet();
    
	      if (pBBLink[link]->busyList.head == NULL)
	        dogStart = 1;
              else
	        dogStart = 0;
  
              /* Add pnode to the busy list */
              listAddTail(&(pBBLink[link]->busyList), pnode);
  
              /* Count the outstanding messages */
              (pBBLink[link]->deviceStatus[pnode->txMsg.node])++;
    
  
              /* If just added something to an empty busy list, start the dog */
              if (dogStart)
              {
		int ticks;
                now = tickGet();
                ticks = pBBLink[link]->busyList.head->retire - now;
                if(ticks<=0) {
	            printf("xvmeWdTask wdStart ticks %d\n",ticks);
	            ticks=1;
                }
                wdStart(pBBLink[link]->watchDogId, ticks, xvmeTmoHandler, link);
              }
	      epicsMutexUnlock(pBBLink[link]->busyList.sem);
	    }

	    /* Tell the 8044 to fire out the message now */
            pBBLink[link]->l.XycomLink.bbRegs->cmnd = BB_SEND_CMD; /* forward it now */
	  }
	  else
	  { /* Aborted transmission operation, re-queue the message */
	    epicsMutexMustLock(pBBLink[link]->queue[BB_Q_HIGH].sem);
	    listAddHead(&(pBBLink[link]->queue[BB_Q_HIGH]), pnode);
	    epicsMutexUnlock(pBBLink[link]->queue[BB_Q_HIGH].sem);
#ifdef BB_SUPER_DEBUG
            BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_ABORT);
#endif
	    if (bbDebug)
	    {
	      printf("Message in progress when abort issued:\n");
	      drvBitBusDumpMsg(&pnode->txMsg);
	    }
	  }
/* BUG -- I don't really need this */
	/* break;*/			/* stop checking the fifo queues */
        }
        else
        { /* we have no xacts that can be processed at this time */
	  epicsMutexUnlock(pBBLink[link]->queue[prio].sem);
        }
     	prio--;				/* look at the next prio queue */
      }
    }
  }
}
/******************************************************************
 *
 * Super debugging facility.
 *
 ******************************************************************/
int
drvBitBusDumpMsg(struct bitBusMsg *pbbMsg)
{
   char	ascBuf[15];	/* for ascii xlation part of the dump */
   int	x;
   int	y;
   int	z;

   printf("Link %d, length %d, route %d, node %d, tasks %d, cmd %d\n",
     pbbMsg->link, pbbMsg->length, pbbMsg->route, pbbMsg->node,
     pbbMsg->tasks, pbbMsg->cmd);

  x = BB_MSG_HEADER_SIZE;
  y = pbbMsg->length;
  z = 0;

  while (x < y)
  {
    printf("%2.2X ", pbbMsg->data[z]);
    ascBuf[z] = pbbMsg->data[z];

    if (!((ascBuf[z] >= 0x20) && (ascBuf[z] <= 0x7e)))
      ascBuf[z] = '.';

    x++;
    z++;
  }

  while (x < BB_MAX_MSG_LENGTH)
  {
    printf("   ");
    x++;
  }

  ascBuf[z] = '\0';
  printf("   *%s*\n", ascBuf);
  return(OK);
}

#ifdef BB_SUPER_DEBUG
/*
 * Place an event into the history buffer.
 */
static int BBSetHistEvent(int link, struct dpvtBitBusHead *pXact, int State)
{
  epicsMutexMustLock(pBBLink[link]->History.sem);


  pBBLink[link]->History.Num++;

  pBBLink[link]->History.Xact[pBBLink[link]->History.Next].Time = tickGet();
  pBBLink[link]->History.Xact[pBBLink[link]->History.Next].State = State;
  if (pXact != NULL)
    memcpy (&pBBLink[link]->History.Xact[pBBLink[link]->History.Next].Xact,
        pXact,  sizeof(struct dpvtBitBusHead));

  if (link==bbDebugLink
  && pBBLink[link]->History.Xact[pBBLink[link]->History.Next].Xact.txMsg.node==bbDebugNode)
    BBDumpXactHistory(&pBBLink[link]->History.Xact[pBBLink[link]->History.Next]);

  if (++pBBLink[link]->History.Next == BB_SUPER_DEBUG_HIST_SIZ)
    pBBLink[link]->History.Next = 0;

  epicsMutexUnlock(pBBLink[link]->History.sem);
  return(0);
}

int BBHistDump(int link)
{
  int	count;
  int	ix;

  if (checkLink(link) == ERROR)
  {
    printf(" Link %d is not valid\n", link);
    return(-1);
  }
  printf("Dumping bitbus history for link %d\n", link);
  epicsMutexMustLock(pBBLink[link]->History.sem);

  if (pBBLink[link]->History.Num < BB_SUPER_DEBUG_HIST_SIZ)
  {
    count = pBBLink[link]->History.Num;
    ix = 0;
  }
  else
  {
    count = BB_SUPER_DEBUG_HIST_SIZ;
    ix = pBBLink[link]->History.Next;
  }

  while(count)
  {
    printf("%8.8d:", pBBLink[link]->History.Xact[ix].Time);

    if (BBDumpXactHistory(&pBBLink[link]->History.Xact[ix]) != 0)
      count = 0;
    else
      --count;

    if (++ix >= BB_SUPER_DEBUG_HIST_SIZ)
      ix = 0;
  }

  epicsMutexUnlock(pBBLink[link]->History.sem);
  return(0);
}
STATIC int BBDumpXactHistory(XactHistStruct *pXact)
{
    switch (pXact->State)
    {
    /* No BB message present */
    case XACT_HIST_STATE_RESET:		
      printf("Link Reset:\n");
      break;
    case XACT_HIST_STATE_RX_ABORT:
      printf("Link RX aborted\n");
      break;
    case XACT_HIST_STATE_TX_ABORT:
      printf("Link TX aborted\n");
      break;
    case XACT_HIST_STATE_RX_URCMD:
      printf("Link RX got unexpected RCMD\n");
      break;

    /* TX message only */
    case XACT_HIST_STATE_QUEUE:
      printf("Message Queued:\n");
      goto do_tx_only;
    case XACT_HIST_STATE_TX:
      printf("Message Transmitted:\n");
      goto do_tx_only;
    case XACT_HIST_STATE_WD_TIMEOUT:
      printf("Watchdog Timeout:\n");
      goto do_tx_only;
    case XACT_HIST_STATE_TX_STUCK:
      printf("Transmitter stuck:\n");
      goto do_tx_only;
    case XACT_HIST_STATE_TX_RESET:
      printf("TX just sent a RAC RESET:\n");

    do_tx_only:
      drvBitBusDumpMsg(&pXact->Xact.txMsg);
      break;

    /* RX message only */
    case XACT_HIST_STATE_RX_UNSOLICITED:
      printf("RX unsolicited message:\n");
      drvBitBusDumpMsg(&pXact->Xact.rxMsg);
      break;

    /* TX and RX messages present */
    case XACT_HIST_STATE_RX:
      printf("RX response... complete transaction:\n");
      drvBitBusDumpMsg(&pXact->Xact.txMsg);
      drvBitBusDumpMsg(&pXact->Xact.rxMsg);
      break;

    default:
      printf("Corrupt bitbus debug state!\n");
      return(-1);
    }
  return(0);
}
#endif
/******************************************************************
 * FUNCTION:    pepRxTask()
 * PURPOSE :    This function is started as a task during driver init
 *              time. It's purpose is to read messages off the receive
 *              FIFO. After a complete message is read, the soliciting
 *              message's completion routines are started by the use
 *              of semaphores and/or a callbackRequest().
 * ARGS IN :    link
 * ARGS OUT:    none
 * GLOBALS:     none
 ******************************************************************/
#define BBRX_HEAD	1
#define	BBRX_DATA	2
#define BBRX_RCMD	3
#define	BBRX_IGN	4
void pepRxTask(int link)
{
  int           rxState;       /* current state of the receiver */

  unsigned char rxHead[BB_MSG_HEADER_SIZE]; /* header of current rx msg */
  unsigned char *rxMsg;         /* where to place next byte (after header) */
  int           rxTCount;       /* byte counter for data in rxHead */
  unsigned char	ch;
  struct dpvtBitBusHead *rxDpvtHead;  /* for message currently receiving */

  int		lockKey;		/* used for intLock calls */
  int		packageComplete;	/* indicates end of package read */

  rxMsg = (unsigned char *) NULL;
  rxState = BBRX_HEAD;
  rxTCount = 0;
  rxDpvtHead = (struct dpvtBitBusHead *) NULL;
  packageComplete = 0;

  while (1) 
  {
    /* Wait for RX interrupts, but only if no chars are ready */
    if ((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & PEP_BB_RFNE) == 0) 
    {
      /* Enable interrupts and check again in case PB-BIT blew it */
      lockKey = intLock();
      pBBLink[link]->l.PepLink.bbRegs->int_vec = pBBLink[link]->IrqVector;
      intUnlock(lockKey);
      
      while (((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & PEP_BB_RFNE) == 0) 
	     && (pBBLink[link]->abortFlag == 0)) 
      {
        /* Re-enable ints here each time in case board got reset */
	lockKey = intLock();
        pBBLink[link]->l.PepLink.bbRegs->int_vec = pBBLink[link]->IrqVector;
	intUnlock(lockKey);
	
	if (bbDebug>35) printf("pepRxTask(%d): waiting on IRQ\n", link);
	/* wait for data bytes */
        epicsEventMustWait(pBBLink[link]->l.PepLink.rxInt);
	if (bbDebug>35) printf("pepRxTask(%d): got IRQ\n", link);
      }
      /* Disable RX Interrupts (prevents unnecessary context switching) */
      lockKey = intLock();
      pBBLink[link]->l.PepLink.bbRegs->int_vec = 0;
      intUnlock(lockKey);
    }

    if (pBBLink[link]->abortFlag == 0) 
    {
      /* READ ONE CHAR FROM RECEIVE FIFO */
      ch = pBBLink[link]->l.PepLink.bbRegs->data;

      /* check to see if we got a data byte or a command byte */
      if ((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & PEP_BB_RCMD) == PEP_BB_RCMD)
	packageComplete = 1;
      
      switch (rxState) {
      case BBRX_HEAD:	/* getting the head of a new message */
	if (rxTCount > 1)  /* Toss the 2 PEP specific header bytes */
	  rxHead[rxTCount] = ch;
	if (bbDebug>21)
	  printf("pepRxTask(%d): >%2.2X< (Header)\n", link, ch);
	
	if (++rxTCount == 7) 
	{
	  /* find the message this is a reply to */
	  /* rxTCount += 2; PEP header bytes already messed up the count (jrw)*/
	  
	  /* Lock the busy list */
	  epicsMutexMustLock(pBBLink[link]->busyList.sem);
	  
	  rxDpvtHead = pBBLink[link]->busyList.head;
	  while (rxDpvtHead != NULL) {
	    if (bbDebug>19)
	      printf("pepRxTask(%d): checking reply against %p\n",
		     link, rxDpvtHead);
	    
	    /* see if node's match */
	    if (rxDpvtHead->txMsg.node == rxHead[4]) {
	      /* see if the tasks match */
	      /* if (rxDpvtHead->txMsg.tasks == rxHead[5]) */
	      {
		/* They match, finish putting response into the rxMsg buffer */
		if (bbDebug>4)
		  printf("pepRxTask(%d): reply to %p\n", 
			 link, rxDpvtHead);
		
		/* Delete the node from the list */
		listDel(&(pBBLink[link]->busyList), rxDpvtHead);
		
		/* If busy list is empty, stop the dog */
		if (pBBLink[link]->busyList.head == NULL)
		  wdCancel(pBBLink[link]->watchDogId);
		
		if (rxHead[6] == 0x91)
		{ /* something bad happened... inject a delay to the */
		  /* requested timeout duration. */

                  if ((link == bbErrorTrigLink) &&
                      ((bbErrorTrigNode < 0) || (bbErrorTrigNode == rxHead[2])))
                      pulseSysMon();

                  /* keep track of timeouts per node */
                  if(rxHead[4] < BB_NUM_NODES) {
                      bbNode91[link][rxHead[4]]++;
                  }

		  if (bbDebug)
		    printf("pepRxTask(%d): 0x91 node %d, start synthetic delay\n",
                        link, rxHead[4]);
		  (pBBLink[link]->syntheticDelay[rxDpvtHead->txMsg.node])
                      = rxDpvtHead->retire;
		  pBBLink[link]->DelayCount++;
		}
		else
		{ /* decrement the number of outstanding messages to the node */
		  (pBBLink[link]->deviceStatus[rxDpvtHead->txMsg.node])--;
		}

		/* Wake up Link Task in case was waiting on "this" node */
		epicsEventSignal(pBBLink[link]->linkEventSem); 
		break;                          /* get out of the while() */
	      }
	    }
	    rxDpvtHead = rxDpvtHead->next; /* Keep looking */
	  }
	  
	  epicsMutexUnlock(pBBLink[link]->busyList.sem);

	  /* Couldn't find a match... fake one so can print the bad message */
	  if (rxDpvtHead == NULL) {
              if (bbDebug)
              {
                 printf("pepRxTask(%d): msg from node %d unsolicited:",
                 link, rxHead[2]);
              }
              rxState = BBRX_IGN;
              break;
          }
	  rxDpvtHead->rxMsg.length = rxHead[2];
	  rxDpvtHead->rxMsg.route = rxHead[3];
	  rxDpvtHead->rxMsg.node = rxHead[4];
	  rxDpvtHead->rxMsg.tasks = rxHead[5];
	  rxDpvtHead->rxMsg.cmd = rxHead[6];
	  rxMsg = rxDpvtHead->rxMsg.data;

	  rxDpvtHead->status = BB_OK;     /* OK, unless BB_LENGTH */
	  rxState = BBRX_DATA;            /* finish reading till RCMD */
	}
	break;
	
      case BBRX_DATA:	/* finish reading data portion of message */
	if (rxTCount >= rxDpvtHead->rxMaxLen) {
	  rxState = BBRX_IGN;	/* toss the rest of the data */
	  rxDpvtHead->status = BB_LENGTH; /* set driver status */
	  if (bbDebug>22)
	    printf("pepRxTask(%d): %2.2X (Ignored)\n", link, ch);
	  
	}
	else {
	  *rxMsg = ch;
	  if (bbDebug>22)
	    printf("pepRxTask(%d): %2.2X (Data)\n", link, ch);
	  rxMsg++;
	  rxTCount++;
	}
	break;
	
      case BBRX_IGN:
	if (bbDebug>22)
	  printf("pepRxTask(%d): %2.2X (Ignored)\n", link, ch);
	break;
      }

      if (packageComplete) 
      {
	if (rxDpvtHead == NULL) 
	{
	  if (bbDebug > 22)
	    printf("pepRxTask(%d): got unexpected completion status\n", link);
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, NULL, XACT_HIST_STATE_RX_URCMD);
#endif
	}
	else 
	{
	  rxDpvtHead->status = BB_OK;
	  if (bbDebug>24)
	    printf("pepRxTask(%d): RX command byte = %2.2X\n", link, ch);

#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, rxDpvtHead, XACT_HIST_STATE_RX);
#endif
	  
          snoopMessage("pepRxTask",link,&rxDpvtHead->rxMsg);
	  if (rxDpvtHead->callback.callback != NULL) 
	  {
	    if (bbDebug>8)
	      printf("pepRxTask(%d): invoking the callbackRequest\n", link);
	    callbackRequest(&rxDpvtHead->callback); /* schedule completion processing */
	  } else if (rxDpvtHead->psyncSem != NULL) {
	      epicsEventSignal(*(rxDpvtHead->psyncSem));
	  }
	}
	/* Reset the state of the RxTask to expect a new message */
	rxMsg = (unsigned char *) NULL;
	rxState = BBRX_HEAD;
	rxTCount = 0;
	rxDpvtHead = (struct dpvtBitBusHead *) NULL;
	packageComplete = 0;
      }
    }
    else 
    {
      /* Link abort state is active reset receiver link status now */
      if (rxDpvtHead != NULL) 
      {
	/* This xact is not on the busy list, put it back on */
	epicsMutexMustLock(pBBLink[link]->busyList.sem);
	(pBBLink[link]->deviceStatus[rxDpvtHead->txMsg.node])++;
	listAddTail(&(pBBLink[link]->busyList), rxDpvtHead);
	epicsMutexUnlock(pBBLink[link]->busyList.sem);
      }
#ifdef BB_SUPER_DEBUG
      BBSetHistEvent(link, rxDpvtHead, XACT_HIST_STATE_RX_ABORT);
#endif

      rxMsg = (unsigned char *) NULL;
      rxState = BBRX_HEAD;
      rxTCount = 0;
      rxDpvtHead = (struct dpvtBitBusHead *) NULL;
      
      /* Tell the watch dog I am ready for the reset (reset in the dog task) */
      pBBLink[link]->rxAbortAck = 1;
      
      if (bbDebug)
        printf("pepRxTask(%d): resetting due to abort status\n", link);
      
      /* wait for link state to become active again */
      while (pBBLink[link]->abortFlag != 0)
	epicsThreadSleep(RESET_POLL_TIME);
      
      if (bbDebug)
        printf("pepRxTask(%d): restarting after abort\n", link);
    }
  }
}

/******************************************************************
 * FUNCTION:    pepWdTask()
 * PURPOSE :    
 *              
 * ARGS IN :    none
 * ARGS OUT:    none
 * GLOBALS:     none
 ******************************************************************/
void pepWdTask(int link)
{
  struct dpvtBitBusHead *pnode;
  struct dpvtBitBusHead *npnode;
  unsigned      long    now;


  pBBLink[link]->nukeEm = 0;		/* Make sure the nuke status is clear */
  while(1) {
    /* SLEEP UNTIL WATCHDOG TIMER ROUTINE GIVES SEMAPHORE */
    epicsEventMustWait(pBBLink[link]->watchDogSem);
    now = tickGet();		/* what time is it? */
    
    if (pBBLink[link]->nukeEm != 0)
      printf("Bitbus manual reset being issued on link %d\n", link);
    
    if (bbDebug>4)
      printf("pepWdTask(%d): (Watchdog) checking busy list\n", link);
    
    pBBLink[link]->rxAbortAck = 0;
    pBBLink[link]->txAbortAck = 0;
    
    if (pBBLink[link]->nukeEm != 0) {
      /* set abort status and wait for the abort acks */
      pBBLink[link]->abortFlag = 1;
      
      /* wake up the Tx task so it can observe the abort status */
      epicsEventSignal(pBBLink[link]->linkEventSem);
      
      /* wake up the Rx task so it can observe the abort status */
      epicsEventSignal(pBBLink[link]->l.PepLink.rxInt);
      
      /* sleep until abort ack from Tx & Rx tasks */
      while ((pBBLink[link]->rxAbortAck == 0) && 
	     (pBBLink[link]->txAbortAck == 0))
	epicsThreadSleep(RESET_POLL_TIME);

#ifdef BB_SUPER_DEBUG
      if (pBBLink[link]->nukeEm < 2)
      { /* Do a history dump since this is not supposed to happen */
        BBHistDump(link);
      }
#endif
    }
    /*
     * Run thru entire busy list to see if there are any transactions
     * that have been waiting on a response for too long a period.
     */
    epicsMutexMustLock(pBBLink[link]->busyList.sem);
    pnode = pBBLink[link]->busyList.head;
    
    while (pnode != NULL) {
      npnode = pnode->next;   /* remember where we were in the list */
      
      if ((pBBLink[link]->nukeEm != 0) || (pnode->retire <= now)) {
	/* Get rid of the request and set error status etc... */
	listDel(&(pBBLink[link]->busyList), pnode);
	
#ifdef BB_SUPER_DEBUG
        BBSetHistEvent(link, pnode, XACT_HIST_STATE_WD_TIMEOUT);
#endif
	if (bbDebug)
	{
	  printf("pepWdTask(%d): TIMEOUT on bitbus message:\n", link);
	  drvBitBusDumpMsg(&pnode->txMsg);
	}

        if(pnode->txMsg.node < BB_NUM_NODES) {
            bbNodeTMO[link][pnode->txMsg.node]++;
        }
	
	/* BUG -- do this here or defer until RX gets a response? */
        (pBBLink[link]->deviceStatus[pnode->txMsg.node])--; /* fix device status */
	pnode->status = BB_TIMEOUT;
	
	/* Make the callbackRequest if one was spec'd */
        snoopMessage("pepWdTask",link,&pnode->txMsg);
	if(pnode->callback.callback != NULL)
	{
	    if (bbDebug>2)
	    {
	      printf("pepWdTask(%d): invoking the callbackRequest %p %d\n",
                  link, pnode->callback.callback, pnode->callback.priority);
	    }
	  callbackRequest(&pnode->callback);     /* schedule completion processing */
	}
	else
	{
	  /* Release a completion lock if one was spec'd */
	  if (pnode->psyncSem != NULL) epicsEventSignal(*(pnode->psyncSem));
	}
      }
      pnode = npnode;
    }
    
    /* Finish the link reboot if necessary */
    if (pBBLink[link]->nukeEm != 0) 
    {
      /* shut down the bitbus card. */
      pepReset(link);

      if (pBBLink[link]->nukeEm == 2)
      {
	/* 
	 * Stop the watchdog task so the link stays dead.
	 * Since the busy list is locked and the nukeEm flag is still set,
	 * this link can not do any more work.
	 */
	abort();
      }

      /* clear the abort_flag */
      pBBLink[link]->abortFlag = 0;
      
      pBBLink[link]->nukeEm = 0;
    }

    if (pBBLink[link]->busyList.head != NULL) {
      int ticks;
      /* Restart the dog timer */
      if (bbDebug>5)
	printf("pepWdTask(%d): restarting watch dog timer\n", link);
      
      ticks = pBBLink[link]->busyList.head->retire - now;
      if(ticks<=0) {
	  printf("pepWdTask wdStart ticks %d\n",ticks);
	  ticks = 1;
      }
      wdStart(pBBLink[link]->watchDogId, ticks, pepTmoHandler, link);
    }

    epicsMutexUnlock(pBBLink[link]->busyList.sem);

    /* In case the TX is waiting on a node that just timed out */
    epicsEventSignal(pBBLink[link]->linkEventSem);
  }
}

/******************************************************************
 * FUNCTION:    pepTxTask()
 * PURPOSE :    This function is started as a task during driver init
 *              time. It's purpose is to keep the link busy with 
 *              outgoing messages. It awaits user-calls to qBBReq()
 *              (new message to send). It also awaits wake up calls
 *              from the Wd timer (a timed out message to a node,
 *              once handled, means Tx is free to send again to that
 *              node).
 *              
 * ARGS IN :    link
 * ARGS OUT:    none
 * GLOBALS:     none
 ******************************************************************/
int pepminstuck = 1000;
void pepTxTask(int link)
{
  struct dpvtBitBusHead	*pnode;
  int			prio;
  int			working;
  int			dogStart;
  int                   stuck;
  
  int			txTCount;
  int			txCCount;
  unsigned char		*txMsg;
  unsigned	long	now;
  
  if (bbDebug>1)
    printf("pepTxTask started for link %d\n", link);

  while(1) {
    if (pBBLink[link]->abortFlag != 0) {
      /* let the dog know we are ready */
      pBBLink[link]->txAbortAck = 1;

#ifdef BB_SUPER_DEBUG
      BBSetHistEvent(link, NULL, XACT_HIST_STATE_TX_ABORT);
#endif
      
      if (bbDebug)
        printf("pepTxTask(%d): resetting due to abort status\n", link);
      
      while (pBBLink[link]->abortFlag != 0)
	epicsThreadSleep(RESET_POLL_TIME);	/* wait for link to reset */
      
      if (bbDebug)
        printf("pepTxTask(%d): restarting after abort\n", link);
    }
    else
    {
      if (pBBLink[link]->DelayCount)
	epicsEventWaitWithTimeout(pBBLink[link]->linkEventSem, 2.0);
      else
	epicsEventMustWait(pBBLink[link]->linkEventSem);
    }

    if (bbDebug>5)
      printf("pepTxTask(%d): got an event\n", link);
    
    working = 1;
    while ((working != 0) && (pBBLink[link]->abortFlag == 0) &&
	   (pBBLink[link]->busyList.elements < PepMaxOutstandMsgs)) {
      working = 0;
      
      prio = BB_NUM_PRIO-1;
      while ((prio >= 0) && (pBBLink[link]->abortFlag == 0)) {
	/* see if the queue has anything in it */
	epicsMutexMustLock(pBBLink[link]->queue[prio].sem);
	
	if ((pnode = pBBLink[link]->queue[prio].head) != NULL) 
	{
	  now = tickGet();
	  while (pBBLink[link]->deviceStatus[pnode->txMsg.node] == BB_BUSY) 
	  {
	    if ((pBBLink[link]->syntheticDelay[pnode->txMsg.node] != 0)
		&& (pBBLink[link]->syntheticDelay[pnode->txMsg.node] < now))
	    {
	      if (bbDebug)
	        printf("pepTxTask(%d): terminating synthetic idle on node %d\n",
                     link, pnode->txMsg.node);
	      (pBBLink[link]->deviceStatus[pnode->txMsg.node])--;
	      pBBLink[link]->syntheticDelay[pnode->txMsg.node] = 0;
	      pBBLink[link]->DelayCount--;
	    }
	    else
	    {
	      if ((pnode = pnode->next) == NULL)
	        break;
	    }
	  }
	}

	if (pnode != NULL) {  /* have an xact to start processing */
	  working = 1;
	  
	  /* delete the node from the inbound fifo queue */
	  listDel(&(pBBLink[link]->queue[prio]), pnode);
	  
	  epicsMutexUnlock(pBBLink[link]->queue[prio].sem);
#ifdef BB_SUPER_DEBUG
          BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX);
#endif

	  
	  if (bbDebug>3)
	    printf("pepTxTask(%d): got xact, pnode=%p\n", link, pnode);
	  
	  /* Send the message in polled mode */
	  txTCount = pnode->txMsg.length;

	  /* BUG -- would be nice if we verify the length >6 here */

	  txCCount = 0;
	  txMsg = &(pnode->txMsg.length);

	  while ((txCCount < txTCount -1) && (pBBLink[link]->abortFlag == 0)) {

	    stuck = 1000;
	    while (((pBBLink[link]->l.PepLink.bbRegs->stat_ctl & PEP_BB_TFNF) 
	    != PEP_BB_TFNF) && (pBBLink[link]->abortFlag == 0) && --stuck)
	    epicsThreadSleep(.05); /*just waste time*/
if(stuck<pepminstuck) pepminstuck = stuck;
	    if (!stuck)
	    {
#ifdef BB_SUPER_DEBUG
              BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_STUCK);
#endif
	      txStuck(link);
	    }
	    else if (txCCount == 0) {  /* first byte is package type */
	      if (bbDebug>30)
		printf("pepTxTask(%d): outputting %2.2X\n",link,0x81);
	      pBBLink[link]->l.PepLink.bbRegs->data = 0x81;
	    }
	    else if (txCCount == 1) {  /* unused 2nd byte of package */
	      if (bbDebug>30)
		printf("pepTxTask(%d): outputting %2.2X\n",link,0x00);
	      pBBLink[link]->l.PepLink.bbRegs->data = 0x00;
	    } else {                 /* regular ol' message byte */
	      pBBLink[link]->l.PepLink.bbRegs->data = *txMsg;
	      if (bbDebug>30)
		printf("pepTxTask(%d): outputting %2.2X\n",link,*txMsg);
	      if (txCCount != 6)
		txMsg++;
	      else
		txMsg = pnode->txMsg.data;
	    }
	    txCCount++;
	  }
	  
          snoopMessage("pepTxTask",link,&pnode->txMsg);
	  if (pBBLink[link]->abortFlag == 0) {

	    /* don't add to busy list if was a RAC_RESET_SLAVE */
	    if ((pnode->txMsg.cmd == RAC_RESET_SLAVE) && (pnode->txMsg.tasks == 0))
	    {
#ifdef BB_SUPER_DEBUG
              BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_RESET);
#endif
	      if (bbDebug)
	        printf("pepTxTask(%d): RAC_RESET_SLAVE sent\n", link);
	      
	      pBBLink[link]->l.PepLink.bbRegs->stat_ctl = *txMsg;
	      if (bbDebug>30)
		printf("pepTxTask(%d): outputting last byte %2.2X\n",
		       link,*txMsg);

	      pnode->status = BB_OK;
	      
	      if (pnode->callback.callback != NULL) {
		if (bbDebug>4)
		  printf("pepTxTask(%d): invoking the callbackRequest\n", 
			 link);
		callbackRequest(&pnode->callback); /* schedule completion processing */
	      }
    	      else
	      {
	        /* If there is a semaphore for synchronous I/O, unlock it */
	        if (pnode->psyncSem != NULL)
		  epicsEventSignal(*(pnode->psyncSem));
	      }
	      epicsThreadSleep(.25);
	    }
	    else
	    {
	      /* Lock the busy list */
	      epicsMutexMustLock(pBBLink[link]->busyList.sem);

	      pBBLink[link]->l.PepLink.bbRegs->stat_ctl = *txMsg;
	      if (bbDebug>30)
		printf("pepTxTask(%d): outputting last byte %2.2X\n",
		       link,*txMsg);
	      
	      /* set the retire time */
	      pnode->retire = tickGet();
	      if (pnode->ageLimit)
	        pnode->retire += pnode->ageLimit;
	      else
		pnode->retire += BB_DEFAULT_RETIRE_TIME * sysClkRateGet();
	      
	      if (pBBLink[link]->busyList.head == NULL)
		dogStart = 1;
	      else
		dogStart = 0;
	      
	      /* Add pnode to the busy list */
	      listAddTail(&(pBBLink[link]->busyList), pnode);
	      
	      /* Count the outstanding messages */
	      (pBBLink[link]->deviceStatus[pnode->txMsg.node])++;
	      
	      
	      /* If something just added to empty busy list, start the dog */
	      if (dogStart) {
		int ticks;
		now = tickGet();
		ticks = pBBLink[link]->busyList.head->retire - now;
		if(ticks<=0) {
		    printf("pepTxTask wdStart ticks %d\n",ticks);
		    ticks = 1;
		}
		wdStart(pBBLink[link]->watchDogId, ticks, pepTmoHandler, link);
	      }
	      epicsMutexUnlock(pBBLink[link]->busyList.sem);
	    }
	  } else {  /* if abortFlag != 0 */
	    /* Aborted transmission operation, re-queue the message */
	    epicsMutexMustLock(pBBLink[link]->queue[BB_Q_HIGH].sem);
	    listAddHead(&(pBBLink[link]->queue[BB_Q_HIGH]), pnode);
	    epicsMutexUnlock(pBBLink[link]->queue[BB_Q_HIGH].sem);
#ifdef BB_SUPER_DEBUG
            BBSetHistEvent(link, pnode, XACT_HIST_STATE_TX_ABORT);
#endif
	  }

	} else {  /* if pnode == NULL */
	  /* we have no xacts that can be processed at this time */
	  epicsMutexUnlock(pBBLink[link]->queue[prio].sem);
	}
	prio--;				/* look at the next prio queue */
      }
    }
  }
}
/******************************************************************
 * FUNCTION:    dumpStat()
 * PURPOSE :    Temporary function. Shows contents of status
 *              register on PB-BIT module corresponding to link.
 * ARGS IN :    link
 * ARGS OUT:    none
 * GLOBALS:     none
 ******************************************************************/
int pepDumpStat(int link)
{
  unsigned char stat_ctl;

  stat_ctl = pBBLink[link]->l.PepLink.bbRegs->stat_ctl;

  printf("stat_ctl reg: %2X\n",stat_ctl);
  return(OK);
}

/* pulseSysMon.c */
/* Function for pulsing a bit on System Monitor Board */
/* TEMPORARY ROUTINE TO FIND BITBUS PROBLEMS*/
typedef struct SysmonStruct {
  char 	                Pad[36];            /*** nF0 - nF17 36 bytes ***/
  unsigned short        SysmonStatusLink;   /*** nF18 ***/
  unsigned short        SysmonDio;          /*** nF19 ***/
  unsigned short        SysmonIntMask;      /*** nF20 ***/
  unsigned short 	SysmonTemperature;  /*** nF21 ***/
  unsigned short        SysmonWatchdog;     /*** nF22 ***/
  unsigned short 	SysmonVXIVector;    /*** nF23 ***/
  unsigned short        SysmonIntVector;    /*** nF24 ***/
  unsigned short        SysmonIRQ1;         /*** nF25 ***/
  unsigned short        SysmonIRQ2;         /*** nF26 ***/
  unsigned short        SysmonIRQ3;         /*** nF27 ***/
  unsigned short        SysmonIRQ4;         /*** nF28 ***/
  unsigned short        SysmonIRQ5;         /*** nF29 ***/
  unsigned short        SysmonIRQ6;         /*** nF30 ***/
  unsigned short        SysmonIRQ7;         /*** nF31 ***/
} SysmonStruct;

int bitbusTriggerWidth = 1500;

long pulseSysMon() {
  SysmonStruct *SysmonBase;
  unsigned short *pReg;
  int i;
  unsigned short probeVal;
  volatile int j = 0;

  if (sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO, (char *)0x8b80, 
			(char **)&(SysmonBase)) == ERROR) {
    printf("can't convert to local address, aborting\n");
    return(1);
  }
  pReg = &(SysmonBase->SysmonDio);
  
  if(bbDebug) printf("Error Trigger\n"); 

  probeVal = 0xffff;
  vxMemProbe((char*) pReg, WRITE, 2, (char*)&probeVal);
  for (i=0 ; i < bitbusTriggerWidth ; i++) {
    j++;
    j--;
    }
  probeVal = 0x0000;
  vxMemProbe((char*) pReg, WRITE, 2, (char*)&probeVal);

  return(0);
}


long bbDump() {
   int i,j;

   for (j=0; j<BB_NUM_LINKS ; j++) { 
       printf("Link %d\n", j);
       for (i=1;i<BB_NUM_NODES; i++) {
           /* print error tally if non-zero */
           if(bbNodeTMO[j][i] || bbNode91[j][i]) {
               printf("    Node %d  - TMO %d  91s %d\n",i,bbNodeTMO[j][i],
                   bbNode91[j][i]);
           }
       }
   }

   return(0);
}

long bbNodeClear() {
   int i,j;

   for (j=0; j<BB_NUM_LINKS ; j++) {
       for (i=1;i<BB_NUM_NODES; i++) {
           bbNodeTMO[j][i] = 0;
           bbNode91[j][i]  = 0;
       }
   }

   return(0);
}


/**************************************************************************
*
* EPICS device support routines to monitor bbNodeTMO[] and bbNode91[] 
*
**************************************************************************/

/* Create the dset for devBoBbNodeStat and devWfBbNodeStat */
static long init_wf_record();
static long init_bo_record();
static long read_91s();
static long read_TMOs();
static long clear_stats();

struct {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       write_bo;
}devBoBbNodeStatClr={
        5,
        NULL,
        NULL,
        init_bo_record,
        NULL,
        clear_stats};

struct {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       read_wf;
}devWfBbNode91s={
        5,
        NULL,
        NULL,
        init_wf_record,
        NULL,
        read_91s};

struct {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       read_wf;
}devWfBbNodeTMOs={
        5,
        NULL,
        NULL,
        init_wf_record,
        NULL,
        read_TMOs};


static long init_wf_record(pwf)
    struct waveformRecord       *pwf;
{


    /* wf.inp must be a VME_IO type such that Cx indicates the bitbus link
     * and Sy represents the desired waveform : S0 = 91's, S1 = TMO's 
    */
    switch (pwf->inp.type) {
    case (VME_IO) :
        if(pwf->inp.value.vmeio.card >= BB_NUM_LINKS) {
            epicsPrintf("%s.INP >> card number invalid\n", pwf->name);
            pwf->pact = 1;  /* don't allow processing */
            return -1;
        }
        if(pwf->ftvl != menuFtypeUSHORT) {
            epicsPrintf("%s.FTVL >> expected USHORT \n", pwf->name);
            pwf->pact = 1;  /* don't allow processing */
            return -1;
        }
        pwf->nord = 0;
        break;
    default :
        epicsPrintf("%s.INP >> expected VME_IO INP field\n", pwf->name);
        pwf->pact = 1;  /* don't allow processing */
        return -1;
    }
    return(0);
}


static long read_91s(pwf)
    struct waveformRecord       *pwf;
{
    long nRequest = BB_NUM_NODES;

    if(pwf->nelm < BB_NUM_NODES) nRequest = pwf->nelm; 

    memcpy( pwf->bptr, (void *)&bbNode91[pwf->inp.value.vmeio.card][0],
            nRequest * 2);

    pwf->nord = nRequest;

    return(0);
}

static long read_TMOs(pwf)
    struct waveformRecord       *pwf;
{
    long nRequest = BB_NUM_NODES;

    if(pwf->nelm < BB_NUM_NODES) nRequest = pwf->nelm;

    memcpy( pwf->bptr, (void *)&bbNodeTMO[pwf->inp.value.vmeio.card][0],
            nRequest * 2);

    pwf->nord = nRequest;

    return(0);
}


static long init_bo_record(pbo)
    struct boRecord       *pbo;
{


    /* bo.out must be a VME_IO type such that Cx indicates the bitbus link
     * that isdesired to be cleared.
    */
    switch (pbo->out.type) {
    case (VME_IO) :
        if(pbo->out.value.vmeio.card >= BB_NUM_LINKS) {
            epicsPrintf("%s.OUT >> card number invalid\n", pbo->name);
            pbo->pact = 1;  /* don't allow processing */
            return -1;
        }
        break;
    default :
        epicsPrintf("%s.OUT >> expected VME_IO type\n", pbo->name);
        pbo->pact = 1;  /* don't allow processing */
        return -1;
    }
    return(0);
}


static long clear_stats(pbo)
    struct boRecord       *pbo;
{
   int i;

   if(pbo->out.value.vmeio.card < BB_NUM_LINKS) {
       for (i=1;i<BB_NUM_NODES; i++) {
           bbNodeTMO[pbo->out.value.vmeio.card][i] = 0;
           bbNode91[pbo->out.value.vmeio.card][i]  = 0;
       }
   }
   else {
       epicsPrintf("%s.OUT >> card number invalid\n", pbo->name);
   }

    return(0);
}

